mod block_path;
pub use block_path::{BlockPath, BlockPathError};
mod sectored_buf;
mod trailered;
/// Code which enables cryptographic operations.
pub mod crypto;

#[cfg(test)]
mod test_helpers;

#[macro_use]
extern crate static_assertions;

#[cfg(test)]
#[macro_use]
extern crate lazy_static;

use brotli::{CompressorWriter, Decompressor};
use btserde::{self, write_to};
use crypto::{
    AsymKeyPub, Ciphertext, Creds, Decrypter, DecrypterExt, Encrypter, EncrypterExt, Hash,
    HashKind, MerkleStream, SecretStream, Sign, Signature, Signer, SymKey, SymKeyKind,
};

use log::error;
use sectored_buf::SectoredBuf;
use serde::{Deserialize, Serialize};
use serde_big_array::BigArray;
use std::{
    collections::{BTreeMap, HashMap},
    convert::{Infallible, TryFrom},
    fmt::{self, Display, Formatter},
    hash::Hash as Hashable,
    io::{self, Read, Seek, SeekFrom, Write},
    ops::{Add, Sub},
    sync::PoisonError,
    time::{Duration, SystemTime},
};
use trailered::Trailered;

#[derive(Debug)]
pub enum Error {
    MissingWritecap,
    Io(std::io::Error),
    Serde(btserde::Error),
    Crypto(crypto::Error),
    IncorrectSize { expected: usize, actual: usize },
    Custom(Box<dyn std::fmt::Debug + Send + Sync>),
}

impl Error {
    fn custom<E: std::fmt::Debug + Send + Sync + 'static>(err: E) -> Error {
        Error::Custom(Box::new(err))
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        match self {
            Error::MissingWritecap => write!(f, "missing writecap"),
            Error::Io(err) => err.fmt(f),
            Error::Serde(err) => err.fmt(f),
            Error::Crypto(err) => err.fmt(f),
            Error::IncorrectSize { expected, actual } => {
                write!(f, "incorrect size {}, expected {}", actual, expected)
            }
            Error::Custom(err) => err.fmt(f),
        }
    }
}

impl std::error::Error for Error {}

impl From<std::io::Error> for Error {
    fn from(err: std::io::Error) -> Self {
        Error::Io(err)
    }
}

impl From<Error> for std::io::Error {
    fn from(err: Error) -> Self {
        io::Error::new(io::ErrorKind::Other, err)
    }
}

impl From<btserde::Error> for Error {
    fn from(err: btserde::Error) -> Self {
        Error::Serde(err)
    }
}

impl From<crypto::Error> for Error {
    fn from(err: crypto::Error) -> Self {
        Error::Crypto(err)
    }
}

impl From<std::num::TryFromIntError> for Error {
    fn from(err: std::num::TryFromIntError) -> Self {
        Error::custom(err)
    }
}

impl<T: std::fmt::Debug> From<PoisonError<T>> for Error {
    fn from(err: PoisonError<T>) -> Self {
        Error::custom(err.to_string())
    }
}

type Result<T> = std::result::Result<T, Error>;

/// TODO: Remove this once the error_chain crate is integrated.
trait BoxInIoErr<T> {
    fn box_err(self) -> std::result::Result<T, io::Error>;
}

impl<T, E: std::error::Error + Send + Sync + 'static> BoxInIoErr<T> for std::result::Result<T, E> {
    fn box_err(self) -> std::result::Result<T, io::Error> {
        self.map_err(|err| io::Error::new(io::ErrorKind::Other, err))
    }
}

/// TODO: Remove this once the error_chain crate is integrated.
trait StrInIoErr<T> {
    fn str_err(self) -> std::result::Result<T, io::Error>;
}

impl<T, E: Display> StrInIoErr<T> for std::result::Result<T, E> {
    fn str_err(self) -> std::result::Result<T, io::Error> {
        self.map_err(|err| io::Error::new(io::ErrorKind::Other, err.to_string()))
    }
}

/// The default sector size to use for new blocks.
const SECTOR_SZ_DEFAULT: usize = 4096;

/// Trait for types which provide read and write access to blocks.
pub trait Block: Read + Write + Seek + MetaAccess {}

// A trait for streams which only allow reads and writes in fixed sized units called sectors.
pub trait Sectored {
    // Returns the size of the sector for this stream.
    fn sector_sz(&self) -> usize;

    fn assert_sector_sz(&self, actual: usize) -> Result<()> {
        let expected = self.sector_sz();
        if expected == actual {
            Ok(())
        } else {
            Err(Error::IncorrectSize { expected, actual })
        }
    }
}

/// A version of the `Write` trait, which allows integrity information to be supplied when flushing.
pub trait WriteInteg: Write {
    fn flush_integ(&mut self, integrity: &[u8]) -> io::Result<()>;
}

trait Decompose<T> {
    fn into_inner(self) -> T;
}

trait TryCompose<T, U: Decompose<T>> {
    type Error;
    fn try_compose(self, inner: T) -> std::result::Result<U, Self::Error>;
}

trait Compose<T, U> {
    fn compose(self, inner: T) -> U;
}

impl<T, U: Decompose<T>, S: TryCompose<T, U, Error = Infallible>> Compose<T, U> for S {
    fn compose(self, inner: T) -> U {
        let result = self.try_compose(inner);
        // Safety: Infallible has no values, so `result` must be `Ok`.
        unsafe { result.unwrap_unchecked() }
    }
}

pub trait MetaAccess {
    fn block_key(&self) -> Result<SymKey>;
    fn add_readcap_for(&mut self, owner: Principal, key: &dyn Encrypter) -> Result<()>;
    /// Returns the integrity value used to protect the contents of the block.
    fn integrity(&self) -> Option<&[u8]>;
    fn set_path(&mut self, path: BlockPath);
}

/// Extensions to the `Read` trait.
trait ReadExt: Read {
    /// Reads repeatedly until one of the following occur:
    ///  1. The given buffer is full.
    ///  2. A call to `read` returns 0.
    ///  3. A call to `read` returns an error.
    /// The number of bytes read is returned. If an error is returned, then no bytes were read.
    fn fill_buf(&mut self, mut dest: &mut [u8]) -> io::Result<usize> {
        let dest_len_start = dest.len();
        while !dest.is_empty() {
            let byte_ct = match self.read(dest) {
                Ok(byte_ct) => byte_ct,
                Err(err) => {
                    if dest_len_start == dest.len() {
                        return Err(err);
                    } else {
                        // We're not allowed to return an error if we've already read from self.
                        error!("an error occurred in fill_buf: {}", err);
                        break;
                    }
                }
            };
            if 0 == byte_ct {
                break;
            }
            dest = &mut dest[byte_ct..];
        }
        Ok(dest_len_start - dest.len())
    }
}

impl<T: Read> ReadExt for T {}

/// This struct contains all of the metadata fields associated with a block, except for its
/// signature. Since this struct implements `Serialize`, this allows for convenient signature
/// calculations.
#[derive(Serialize, Deserialize, Debug, PartialEq, Clone)]
pub struct BlockMetaBody {
    path: BlockPath,
    readcaps: BTreeMap<Principal, Ciphertext<SymKey>>,
    writecap: Option<Writecap>,
    /// A hash which provides integrity for the contents of the block body.
    integrity: Option<Hash>,
    /// The public key that corresponds to the private key used to sign this header.
    signing_key: AsymKeyPub<Sign>,
}

impl BlockMetaBody {
    fn new<C: Creds>(creds: &C) -> BlockMetaBody {
        BlockMetaBody {
            path: BlockPath::default(),
            readcaps: BTreeMap::new(),
            writecap: creds.writecap().map(|e| e.to_owned()),
            integrity: None,
            signing_key: creds.public_sign().to_owned(),
        }
    }
}

/// Signed metadata associated with a block.
#[derive(Serialize, Deserialize)]
struct BlockMeta {
    body: BlockMetaBody,
    sig: Signature,
}

impl BlockMeta {
    fn new<C: Creds>(creds: &C) -> BlockMeta {
        let body = BlockMetaBody::new(creds);
        let sig = Signature::empty(body.signing_key.scheme());
        BlockMeta { body, sig }
    }
}

struct BlockStream<T, C> {
    trailered: Trailered<T, BlockMeta>,
    meta: BlockMeta,
    meta_body_buf: Vec<u8>,
    creds: C,
}

impl<T: Read + Seek, C: Creds> BlockStream<T, C> {
    fn new(inner: T, creds: C) -> Result<BlockStream<T, C>> {
        let (trailered, trailer) = Trailered::<_, BlockMeta>::new(inner)?;
        let trailer = match trailer {
            Some(mut trailer) => {
                trailer.assert_valid()?;
                // We need to use the writecap and signing_key provided by the current credentials.
                trailer.body.writecap = creds.writecap().map(|e| e.to_owned());
                trailer.body.signing_key = creds.public_sign().to_owned();
                trailer
            }
            None => {
                let mut meta = BlockMeta::new(&creds);
                let block_key = SymKey::generate(SymKeyKind::default())?;
                meta.body
                    .readcaps
                    .insert(creds.principal(), creds.ser_encrypt(&block_key)?);
                meta.body.writecap = creds.writecap().map(|e| e.to_owned());
                meta
            }
        };
        Ok(BlockStream {
            trailered,
            meta: trailer,
            meta_body_buf: Vec::new(),
            creds,
        })
    }
}

impl<T: Write + Seek, C: Signer> Write for BlockStream<T, C> {
    fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
        self.trailered.write(buf)
    }

    fn flush(&mut self) -> io::Result<()> {
        Err(io::Error::new(
            io::ErrorKind::Unsupported,
            "flush is not supported, use flush_integ instead",
        ))
    }
}

impl<T: Write + Seek, C: Signer> WriteInteg for BlockStream<T, C> {
    fn flush_integ(&mut self, integrity: &[u8]) -> io::Result<()> {
        let meta_body = &mut self.meta.body;
        let integ = meta_body.integrity.get_or_insert_with(Hash::default);
        integ.as_mut().copy_from_slice(integrity);
        self.meta_body_buf.clear();
        write_to(&meta_body, &mut self.meta_body_buf).box_err()?;
        self.meta.sig = self
            .creds
            .sign(std::iter::once(self.meta_body_buf.as_slice()))?;
        self.trailered.flush(&self.meta)?;
        Ok(())
    }
}

impl<T: Read + Seek, C> Read for BlockStream<T, C> {
    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
        self.trailered.read(buf)
    }
}

impl<T: Seek, C> Seek for BlockStream<T, C> {
    /// Seeks to the given position in the stream. If a position beyond the end of the stream is
    /// specified, the the seek will be to the end of the stream.
    fn seek(&mut self, pos: SeekFrom) -> io::Result<u64> {
        self.trailered.seek(pos)
    }
}

impl<T, C: Decrypter + Principaled> MetaAccess for BlockStream<T, C> {
    fn block_key(&self) -> Result<SymKey> {
        let readcap = self
            .meta
            .body
            .readcaps
            .get(&self.creds.principal())
            .ok_or(Error::Crypto(crypto::Error::NoReadCap))?;
        Ok(self.creds.ser_decrypt(readcap)?)
    }

    fn add_readcap_for(&mut self, owner: Principal, key: &dyn Encrypter) -> Result<()> {
        let block_key = self.block_key()?;
        let readcap = key.ser_encrypt(&block_key)?;
        self.meta.body.readcaps.insert(owner, readcap);
        Ok(())
    }

    fn integrity(&self) -> Option<&[u8]> {
        self.meta.body.integrity.as_ref().map(|hash| hash.as_ref())
    }

    fn set_path(&mut self, path: BlockPath) {
        self.meta.body.path = path;
    }
}

pub struct BlockOpenOptions<T, C> {
    inner: T,
    creds: C,
    encrypt: bool,
    compress: bool,
}

impl BlockOpenOptions<(), ()> {
    pub fn new() -> BlockOpenOptions<(), ()> {
        BlockOpenOptions {
            inner: (),
            creds: (),
            encrypt: true,
            compress: true,
        }
    }
}

impl<T, C> BlockOpenOptions<T, C> {
    pub fn with_inner<U>(self, inner: U) -> BlockOpenOptions<U, C> {
        BlockOpenOptions {
            inner,
            creds: self.creds,
            encrypt: self.encrypt,
            compress: self.compress,
        }
    }

    pub fn with_creds<D>(self, creds: D) -> BlockOpenOptions<T, D> {
        BlockOpenOptions {
            inner: self.inner,
            creds,
            encrypt: self.encrypt,
            compress: self.compress,
        }
    }

    pub fn with_encrypt(mut self, encrypt: bool) -> Self {
        self.encrypt = encrypt;
        self
    }

    pub fn with_compress(mut self, compress: bool) -> Self {
        self.compress = compress;
        self
    }
}

impl<T: Read + Write + Seek + 'static, C: Creds + 'static> BlockOpenOptions<T, C> {
    pub fn open(self) -> Result<Box<dyn Block>> {
        let stream = BlockStream::new(self.inner, self.creds)?;
        let block_key = stream.block_key()?;
        let mut stream = MerkleStream::new(stream)?;
        stream.assert_root_integrity()?;
        if self.encrypt {
            let stream = SecretStream::new(block_key).try_compose(stream)?;
            let stream = SectoredBuf::new().try_compose(stream)?;
            Ok(Box::new(stream))
        } else {
            let stream = SectoredBuf::new().try_compose(stream)?;
            Ok(Box::new(stream))
        }
    }
}

impl Default for BlockOpenOptions<(), ()> {
    fn default() -> Self {
        Self::new()
    }
}

impl<T: Write> Decompose<T> for CompressorWriter<T> {
    fn into_inner(self) -> T {
        self.into_inner()
    }
}

impl<T: Read> Decompose<T> for Decompressor<T> {
    fn into_inner(self) -> T {
        self.into_inner()
    }
}

#[derive(Clone)]
pub struct BrotliParams {
    buf_sz: usize,
    quality: u32,
    window_sz: u32,
}

impl BrotliParams {
    pub fn new(buf_sz: usize, quality: u32, window_sz: u32) -> BrotliParams {
        BrotliParams {
            buf_sz,
            quality,
            window_sz,
        }
    }
}

impl<T: Write> TryCompose<T, CompressorWriter<T>> for BrotliParams {
    type Error = Error;
    fn try_compose(self, inner: T) -> Result<CompressorWriter<T>> {
        Ok(CompressorWriter::new(
            inner,
            self.buf_sz,
            self.quality,
            self.window_sz,
        ))
    }
}

impl<T: Read> TryCompose<T, Decompressor<T>> for BrotliParams {
    type Error = Error;
    fn try_compose(self, inner: T) -> Result<Decompressor<T>> {
        Ok(Decompressor::new(inner, self.buf_sz))
    }
}

/// An envelopment of a key, which is tagged with the principal who the key is meant for.
#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub struct Readcap {
    /// The principal this `Readcap` was issued to.
    issued_to: Principal,
    /// An encipherment of a block key using the public key of the principal.
    key: Ciphertext<SymKey>,
}

impl Readcap {
    pub fn new(issued_to: Hash, key: Ciphertext<SymKey>) -> Readcap {
        Readcap {
            issued_to: Principal(issued_to),
            key,
        }
    }
}

#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub struct WritecapBody {
    /// The principal this `Writecap` was issued to.
    issued_to: Principal,
    /// The path where this write caps's validity begins.
    path: BlockPath,
    /// The point in time after which this write cap is no longer valid.
    expires: Epoch,
    /// The public key used to sign this write cap.
    signing_key: AsymKeyPub<Sign>,
}

/// Verifies that a principal is authorized to write blocks in a tree.
#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub struct Writecap {
    /// A container for the fields of this writecap which are covered by the signature.
    body: WritecapBody,
    /// A digital signature which covers all of the fields in the write cap except for next.
    signature: Signature,
    /// The next write cap in the chain leading back to the root.
    next: Option<Box<Writecap>>,
}

/// Fragments are created from blocks using Erasure Encoding and stored with other nodes in the
/// network to provide availability and redundancy of data.
#[derive(Debug, PartialEq, Serialize, Deserialize)]
pub struct Fragment {
    /// The path to the block this fragment is from.
    path: BlockPath,
    /// The serial number of this fragment.
    serial: FragmentSerial,
    /// The actual data.
    body: Vec<u8>,
}

impl Fragment {
    /// Create a new fragment with the given fields. If `path_str` cannot be parsed then a failed
    /// `Result` is returned containing a `PathError`.
    pub fn new(
        path_str: &str,
        serial_num: u32,
        body: Vec<u8>,
    ) -> std::result::Result<Fragment, BlockPathError> {
        let result = BlockPath::try_from(path_str);
        Ok(Fragment {
            path: result?,
            serial: FragmentSerial(serial_num),
            body,
        })
    }
}

/// The body of every non-leaf node in a tree contains this data structure.
#[derive(Debug, PartialEq, Serialize, Deserialize)]
pub struct Directory {
    /// The nodes that are attached to the tree at this block.
    nodes: Vec<Principal>,
    /// This block's descendants.
    children: HashMap<String, HashMap<FragmentSerial, FragmentRecord>>,
}

/// Keeps track of which principal is storing a fragment.
#[derive(Debug, PartialEq, Serialize, Deserialize)]
pub struct FragmentRecord {
    /// The fragment serial number this record is for.
    serial: FragmentSerial,
    /// The principal who is storing this fragment.
    stored_by: Principal,
}

impl FragmentRecord {
    /// Creates a new `FragmentRecord` whose `serial` and `stored_by` fields are set to
    /// the given values.
    pub fn new(serial: u32, stored_by: Hash) -> FragmentRecord {
        FragmentRecord {
            serial: FragmentSerial(serial),
            stored_by: Principal(stored_by),
        }
    }
}

/// An identifier for a security principal, which is any entity that can be authenticated.
#[derive(
    Debug, PartialEq, Eq, Serialize, Deserialize, Hashable, Clone, Default, PartialOrd, Ord,
)]
pub struct Principal(Hash);

impl Principal {
    fn kind(&self) -> HashKind {
        HashKind::from(&self.0)
    }
}

impl Display for Principal {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

/// Trait for types which are owned by a `Principal`.
pub trait Principaled {
    /// Returns the `Principal` that owns `self`, using the given hash algorithm.
    fn principal_of_kind(&self, kind: HashKind) -> Principal;

    /// Returns the `Principal` that owns `self`, using the default hash algorithm.
    fn principal(&self) -> Principal {
        self.principal_of_kind(HashKind::default())
    }
}

/// An instant in time represented by the number of seconds since January 1st 1970, 00:00:00 UTC.
#[derive(Debug, Serialize, Deserialize, Clone, PartialEq, Eq, PartialOrd, Ord, Default)]
pub struct Epoch(u64);

impl Epoch {
    /// Returns the current epoch time.
    fn now() -> Epoch {
        let now = SystemTime::now();
        // If the system clock is before the unix epoch, just panic.
        let epoch = now.duration_since(SystemTime::UNIX_EPOCH).unwrap();
        Epoch(epoch.as_secs())
    }
}

impl Copy for Epoch {}

impl Add<Duration> for Epoch {
    type Output = Self;
    fn add(self, other: Duration) -> Self {
        Epoch(self.0 + other.as_secs())
    }
}

impl Sub<Duration> for Epoch {
    type Output = Self;
    fn sub(self, other: Duration) -> Self {
        Epoch(self.0 - other.as_secs())
    }
}

/// The serial number of a block fragment.
#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Hashable)]
struct FragmentSerial(u32);

#[cfg(test)]
mod tests {
    use std::{fs::OpenOptions, io::Cursor, path::PathBuf};

    use crate::crypto::{
        tpm::{TpmCredStore, TpmCreds},
        ConcreteCreds, CredStore, CredsPriv,
    };

    use super::*;
    use tempdir::TempDir;
    use test_helpers::*;

    #[test]
    fn brotli_compress_decompress() {
        const SECT_SZ: usize = SECTOR_SZ_DEFAULT;
        const SECT_CT: usize = 16;
        let params = BrotliParams::new(SECT_SZ, 8, 20);
        let mut memory = Cursor::new([0u8; SECT_SZ * SECT_CT]);
        {
            let write: CompressorWriter<_> = params
                .clone()
                .try_compose(&mut memory)
                .expect("compose for write failed");
            write_fill(write, SECT_SZ, SECT_CT);
        }
        memory.seek(SeekFrom::Start(0)).expect("seek failed");
        {
            let read: Decompressor<_> = params
                .try_compose(&mut memory)
                .expect("compose for read failed");
            read_check(read, SECT_SZ, SECT_CT);
        }
    }

    #[test]
    fn block_can_create_empty() {
        let harness = SwtpmHarness::new().expect("failed to start swtpm");
        let context = harness.context().expect("failed to retrieve context");
        let cred_store = TpmCredStore::new(context, harness.state_path())
            .expect("failed to create TpmCredStore");
        let creds = cred_store.node_creds().expect("failed to get node creds");
        BlockOpenOptions::new()
            .with_inner(Cursor::new(Vec::<u8>::new()))
            .with_creds(creds)
            .with_encrypt(true)
            .open()
            .expect("failed to open block");
    }

    struct BlockTestCase {
        root_creds: TpmCreds,
        node_creds: TpmCreds,
        _swtpm: SwtpmHarness,
        temp_dir: TempDir,
    }

    impl BlockTestCase {
        const ROOT_PASSWORD: &'static str = "(1337Prestidigitation7331)";

        fn new() -> BlockTestCase {
            let temp_dir = TempDir::new("block_test").expect("failed to create temp dir");
            let swtpm = SwtpmHarness::new().expect("failed to start swtpm");
            let context = swtpm.context().expect("failed to retrieve context");
            let cred_store = TpmCredStore::new(context, swtpm.state_path())
                .expect("failed to create TpmCredStore");
            let root_creds = cred_store
                .gen_root_creds(Self::ROOT_PASSWORD)
                .expect("failed to get root creds");
            let mut node_creds = cred_store.node_creds().expect("failed to get node creds");
            let writecap = root_creds
                .issue_writecap(
                    node_creds.principal(),
                    vec!["nodes".to_string(), "phone".to_string()],
                    Epoch::now() + Duration::from_secs(3600),
                )
                .expect("failed to issue writecap");
            node_creds.set_writecap(writecap);
            BlockTestCase {
                temp_dir,
                _swtpm: swtpm,
                node_creds,
                root_creds,
            }
        }

        fn fs_path(&self, path: &crate::BlockPath) -> PathBuf {
            let mut fs_path = self.temp_dir.path().to_owned();
            fs_path.extend(path.components());
            fs_path
        }

        fn open_new(&mut self, path: &crate::BlockPath) -> Box<dyn Block> {
            let file = OpenOptions::new()
                .create_new(true)
                .read(true)
                .write(true)
                .open(&self.fs_path(path))
                .expect("failed to open file");
            let mut block = BlockOpenOptions::new()
                .with_inner(file)
                .with_creds(self.node_creds.clone())
                .with_encrypt(true)
                .open()
                .expect("failed to open block");
            block.set_path(self.node_creds.writecap().unwrap().body.path.clone());
            block
        }

        fn open_existing(&mut self, path: &crate::BlockPath) -> Box<dyn Block> {
            let file = OpenOptions::new()
                .read(true)
                .write(true)
                .open(&self.fs_path(path))
                .expect("failed to reopen file");
            BlockOpenOptions::new()
                .with_inner(file)
                .with_creds(self.node_creds.clone())
                .with_encrypt(true)
                .open()
                .expect("failed to reopen block")
        }
    }

    #[test]
    fn block_contents_persisted() {
        const EXPECTED: &[u8] = b"Silly sordid sulking sultans.";

        let mut case = BlockTestCase::new();
        let path = BlockPath::new(case.root_creds.principal(), vec!["test.blk".to_string()]);
        {
            let mut block = case.open_new(&path);
            block.write(EXPECTED).expect("failed to write");
            block.flush().expect("flush failed");
        }
        let mut block = case.open_existing(&path);
        let mut actual = [0u8; EXPECTED.len()];
        block.read(&mut actual).expect("read failed");
        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn block_write_twice() {
        const EXPECTED: &[u8] = b"Cool callous calamitous colonels.";
        const MID: usize = EXPECTED.len() / 2;

        let mut case = BlockTestCase::new();
        let path = BlockPath::new(case.root_creds.principal(), vec!["test.blk".to_string()]);
        {
            let mut block = case.open_new(&path);
            block.write(&EXPECTED[..MID]).expect("first write failed");
            block.flush().expect("first flush failed");
        }
        {
            let mut block = case.open_existing(&path);
            block
                .seek(SeekFrom::Start(MID.try_into().unwrap()))
                .expect("seek failed");
            block.write(&EXPECTED[MID..]).expect("second write failed");
            block.flush().expect("second flush failed");
        }
        {
            let mut block = case.open_existing(&path);
            let mut actual = [0u8; EXPECTED.len()];
            block.read(&mut actual).expect("read failed");
            assert_eq!(EXPECTED, actual);
        }
    }

    #[test]
    fn block_write_with_different_creds() {
        const EXPECTED: &[u8] = b"Cool callous calamitous colonels.";
        const MID: usize = EXPECTED.len() / 2;

        let mut case = BlockTestCase::new();
        let path = BlockPath::new(case.root_creds.principal(), vec!["test.blk".to_string()]);
        let app_creds = {
            let mut app_creds = ConcreteCreds::generate().expect("failed to generate app creds");
            let writecap = case
                .root_creds
                .issue_writecap(
                    app_creds.principal(),
                    path.components().map(|e| e.to_string()).collect(),
                    Epoch::now() + Duration::from_secs(60),
                )
                .expect("failed to issue writecap");
            app_creds.set_writecap(writecap);
            app_creds
        };
        {
            let mut block = case.open_new(&path);
            block
                .add_readcap_for(app_creds.principal(), &app_creds)
                .expect("failed to add readcap");
            block.write(&EXPECTED[..MID]).expect("first write failed");
            block.flush().expect("first flush failed");
        }
        {
            let file = OpenOptions::new()
                .read(true)
                .write(true)
                .open(case.fs_path(&path))
                .expect("failed to reopen file");
            let mut block = BlockOpenOptions::new()
                .with_inner(file)
                // Note that this write is performed using app_creds.
                .with_creds(app_creds)
                .with_encrypt(true)
                .open()
                .expect("failed to reopen block");
            block
                .seek(SeekFrom::Start(MID.try_into().unwrap()))
                .expect("seek failed");
            block.write(&EXPECTED[MID..]).expect("second write failed");
            block.flush().expect("second flush failed");
        }
        {
            let file = OpenOptions::new()
                .read(true)
                .write(true)
                .open(case.fs_path(&path))
                .expect("failed to reopen file");
            let mut block = BlockOpenOptions::new()
                .with_inner(file)
                .with_creds(case.node_creds)
                .with_encrypt(true)
                .open()
                .expect("failed to reopen block");
            let mut actual = [0u8; EXPECTED.len()];
            block.read(&mut actual).expect("read failed");
            assert_eq!(EXPECTED, actual);
        }
    }
}
